# breitensuche.scm
# Breitensuche am Rucksackproblem unter Beachtung der Kohaesion
# Container sind tiefe Listen
### bei dieser Version wird nicht mit einer Warteschlange gearbeitet
### da die Verwaltung dann einfacher ist, wird jeweils die ganze Ebene expandiert

# ----- Container sind Listen mit dem fassungsvermoegen am Kopf -----
def erzeugeContainer(fassungsvermoegen):
    return [fassungsvermoegen, []]

# ----- Liste der vorhandenen Stuecke -----
stuecke= [40,30,30,30,25,25,25,20,20,20,15,10,10,10,10,5]

### ===== ANFANG HILFSFUNKTIONEN =====
def cons(element,liste): return [element]+liste

# ----- fuelle -----
# berechnet die gesamte Fuellung des Containers
def fuellung(container):
    return sum(container[1])
# Test: print( fuellung([100,[30,30]]) )

# ----- fuelle_ein -----
# fuellt ein stueck ein
# die Zulässigkeit muss extern gepfüft werden
def fuelle_ein(stueck, container):
    return [container[0], cons(stueck,container[1])]
# Test: print( fuelle_ein( 20, [100,[30,30]]) )

# ----- Praedikate -----
# Im Sinne der Kohaesion sollten diese Prädikate ausgegliedert werden:
# ----- exakt_voll -----
# exakt_voll prueft, ob alle Inhalte zusammen die fassungsvermoegen ergeben
def exakt_voll(container):
    return fuellung(container) == container[0]
# Test: print( exakt_voll([100,[30,30,20,20]]) )

# ----- zu_voll -----
# zu_voll prueft, ob das Stueck zusammen mit allen Inhalten die fassungsvermoegen ergeben
def zu_voll(stueck, container):
    return (stueck + fuellung(container)) > container[0]
# Test: print( zu_voll(10, [100,[30,30,20,20]]) )

### ===== ENDE HILFSFUNKTIONEN =====

### ===== Weitere Modellierung =====
# Durch die Vorgabe der Funktionen zu_voll und fuelle_ein, bei denen vorher
# geprüft wird, muss das beim Generieren der Elemente der neuen Ebene erfolgen
# der Erfolgsfall lässt sich hier nicht gut zurückgeben, daher sollte die Testung
# in der Steuerung erfolgen
# Dann laesst die Aufgabe sich global in drei Teilaufgaben zerlegen:
# - die Expansion der zulaessigen Nachfolger einer neuen Ebene
# - die Steuerung des Ablaufs 
# - die Aufrufhuelle
# Mit dieser Gliederung wird begonnen:

# ----- Expansion einer Ebene -----
# Wegen der Endrekursion wird eine Akkumulatorliste benoetigt.
# Für die nächste Expansionsebene werden zunächst alle Alternativen
# der bisherigen Ebene (in den akku) übernommen, da sie die Variante
# darstellen, bei denen das aktuelle Stueck nicht eingefügt wird.
# Zusätzlich werden alle zulässigen Moeglichkeiten mit dem neuen Stueck
# generiert und ebenfalls in den akku übernommen.
# Passend zu Python wird hier nicht rekursiv gearbeitet.
# ----- expandiere -----
def expandiere(stueck, alternativen):
    ##input((stueck, alternativen))
    akku=[]
    for alt in alternativen:
        if not zu_voll(stueck, alt):
            akku.append(fuelle_ein(stueck, alt))
    return akku+alternativen

# ----- loesung_gefunden -----
# sucht in den alternativen nach einem container, der exakt voll ist
# gibt den zurück oder #f
### [Hinweis: kein Prädikat, da zwar False zurück gegeben werden kann,
###  aber nicht True zurück gegeben wird, sondern die Lösung]
# Wenn 'optimal' gesucht wird und nicht exakt voll, muss hier
# modifiziert werden.
def loesung_gefunden(alternativen):
    ##input(alternativen)
    for alt in alternativen:
        if exakt_voll(alt):
            return alt
    return False

# ----- Steuerung -----
# steuert die anderen Funktionen
# - ist die Liste stuecke leer, gab es keine Loesung
# - gibt loesung_gefunden die Loesung, wird sie zurueckgegeben
# - sonst wird eine neuer Durchlauf gestartet.
def steuerung(stuecke, alternativen):
##    input((stuecke, alternativen))
##    print()
    gefunden=loesung_gefunden(alternativen)
    if gefunden:            # erst Erfolgsfall testen
        return gefunden
    elif len(stuecke)==0:   # und Misserfolg
        return False
    else:
        return steuerung(stuecke[1:],
                         expandiere(stuecke[0], alternativen))

# ===== Definition und Einbau einer Aufrufhuelle =====
# die Stuecke und der Container werden übergeben

def breitensuche(stuecke, fassungsvermoegen):
    print("breitensuche mit ",
          stuecke,
          " Kapazitaet ",
          fassungsvermoegen,
          " ergibt:\n",
          steuerung(stuecke, [erzeugeContainer(fassungsvermoegen)]), '\n')

# Testaufrufe:
breitensuche(stuecke,100)
breitensuche(stuecke,170)
breitensuche(stuecke,sum(stuecke))
breitensuche(stuecke,33)
breitensuche(stuecke,1000)
breitensuche([70,60,50,40,30,20,20], 260)
##steuerung(stuecke, [erzeugeContainer(170)])
